"""
An accompanying python module for the lecture Scientific Computing II (SS 2016).

This module contains auxiliary routines and sample solutions for the programming tasks which had to be solved by the students during the course.
"""

import os.path
import numpy as np
import scipy.linalg
import matplotlib.pyplot as plt
import numpy.linalg as la
from scipy.spatial.distance import cdist
from scipy.spatial.distance import squareform
from scipy.linalg import solve

from scipy.linalg import cholesky
from scipy.linalg import solve_triangular

from skimage.io import imread

###########################################################################
# data generation
###########################################################################

def toy(x):
    """
    A toy function to explore kernel interpolation and regression.

    The toy function is a superposition of three bell curves and sine function.
    """
    return np.exp(-100*np.power(x-0.5,2)) +\
           0.5*np.exp(-10*np.power(x-0.2,2)) +\
           0.5*np.exp(-100*np.power(x-0.9,2)) +\
           0.1*np.sin(-100*np.power(x-0.5,2))

X_full   = np.linspace(0,1,1000)
X_sample = np.linspace(0,1,10)

heavy_noise = [ 0.31006098, -0.06900007,  0.49688356,  1.20370015,  0.2254434,  -0.15198685,
 -0.93082213,  0.16441064,  0.87895284, -0.52823194]

toy_observed = toy(X_sample)
toy_noisy = toy_observed + heavy_noise

def random_image_data(shape):
    """
    Generates feature vectors randomly sampled from natural image scenes.
    
    Parameters:
      N - number of feature vectors
      d - number of features per vector
    
    Returns:
      data - numpy.ndarray with shape==(N,d)
    
        
    The images are taken from 'http://www.cs.helsinki.fi/u/phoyer/NCimages'. 
    This is the source used in
    
      Bingham, E., & Mannila, H. (2001). Random projection in dimensionality 
      reduction: applications to image and text data. In Proceedings of the 
      seventh ACM SIGKDD international conference on Knowledge discovery and 
      data mining (pp. 245-250). ACM.
    
    """
    # N: number of feature vectors
    # d: number of features per vector
    N, d = shape
    
    n_images = 13 # we have 13 image files
    
    impatch_height = np.floor(np.sqrt(d))
    impatch_width  = np.floor(np.sqrt(d))
    if (impatch_height > 256) |(impatch_width > 256):
        return ValueError('patch size must be less than 256^2')

    N_per_image = N//n_images*np.ones(n_images)
    r = N % n_images
    N_per_image[0:r] = N_per_image[0:r] + np.ones(r)
    N_per_image = N_per_image[np.nonzero(N_per_image)]

    data = np.zeros((N, impatch_width*impatch_height))
    for i in range(np.size(N_per_image)):
        impath = os.path.join(os.path.dirname(__file__),'images/%d.tiff' % (i+1))
        im = imread(impath,as_grey=True)
        im_height, im_width = im.shape
  
        for j in range(int(N_per_image[i])):
            j_offset = np.sum(N_per_image[0:i])
            height_offset = np.random.randint(0,im_height-impatch_height)
            width_offset  = np.random.randint(0,im_width-impatch_width)
            impatch = im[height_offset:height_offset+impatch_height,width_offset:width_offset+impatch_height]
            data[j_offset+j,:] = np.reshape(impatch,[1,impatch_height*impatch_width])
    return data

###########################################################################
# kernels
###########################################################################

from sklearn.metrics.pairwise import rbf_kernel

def gauss_kernel(X, Y, gamma):
    """
    Evaluates the Gauss kernel k(x,y) = exp(-gamma \|x-y\|_2^2) with width gamma
    in all tuples (x,y), x in X, y in Y.
    
    Parameters:
        X - array-like with N elements
        Y - array-like with M elements
        gamma - float
        
    Returns:
        keval - matrix or ndarray of shape (N,M)
    """
    N = np.size(X); M = np.size(Y)
    return rbf_kernel(np.reshape(X, (N,1)), np.reshape(Y, (M,1)),gamma)

def dirichlet(x,m):
    values = np.sin((m+0.5)*x)/np.sin(0.5*x)
    NaNs = np.isnan(values)
    values[NaNs] = 2.0*m+1.0
    return values

def dirichlet_kernel(X, Y, m):
    """
    Evaluates the Gauss kernel Dirichlet kernel D_m(x,y) = sin((m+0.5)*(x-y))/sin(0.5*(x-y)
    in all tuples (x,y), x in X, y in Y.
    
    Parameters:
        X - array-like with N elements
        Y - array-like with M elements
        gamma - float
        
    Returns:
        keval - matrix or ndarray of shape (N,M)
    """
    N = np.size(X); M = np.size(Y)
    dist_matrix = cdist(np.reshape(X, (N,1)), np.reshape(Y, (M,1)),'cityblock')
    return dirichlet(dist_matrix, m)

###########################################################################
# learning algorithms
###########################################################################

def eigenvalue_decomposition(kernelmatrix):
    """
    Computes the eigenvalue decomposition for the given symmetric kernelmatrix.
    
    Parameters:
        kernelmatrix - array-like of shape (N,N)
        
    Returns:
        eigval -  array-like with N elements; the eigenvalues
        Q - matrix of shape (N,N); the normalized (unit “length”) eigenvectors,
            such that the column Q[:,i] is the eigenvector corresponding
            to the eigenvalue eigval[i].
    """
    return la.eigh(np.asmatrix(kernelmatrix))


def compute_predictor(kernel, eigval, Q, sample_points, sample_values, regparam):
    """
    Computes the function approximation obtained via regularized empirical risk minimization.
    
    Parameters:
        kernel - function object which takes two array-like arguments X and Y
        eigval - array-like with N elements; the eigenvalues
        Q      - matrix of shape (N,N); the normalized (unit “length”) eigenvectors,
                 such that the column Q[:,i] is the eigenvector corresponding
                 to the eigenvalue eigval[i].
        sample_points - given sampling points; array-like with N elements
        sample_values - values observed at the given sampling points
        regparam - float; the regularization parameter
        
    Returns:
       predictor - function object which takes an array-like argument X; predictor(X) returns
                   an array-like object containing the values of approximant
                   at the points X
    """
    N = sample_values.size
    diagterms = 1.0/(regparam + eigval)
    coefficients = sample_values*Q*np.diag(diagterms)*Q.transpose()
    return lambda x: np.squeeze(np.asarray(coefficients*kernel(sample_points, x)))

def effective_degrees_of_freedom(eigval, Q):
    min_regparam = 0.0001
    max_regparam = np.max(eigval)*(np.size(eigval)-3)/3 # crude upper estimate for edf=3, the lowest useful edf
    
    list_of_regparams = []
    list_of_edfs      = []
    do_search = True
    
    edf = 3
    while do_search:
        trace = lambda x: np.sum(eigval/(x+eigval))-edf
        try:
            regp = scipy.optimize.brentq(trace,min_regparam,max_regparam)
            list_of_regparams.append(regp)
            list_of_edfs.append(edf)
            edf = edf+1
        except ValueError:
            do_search = False
    return list_of_edfs, list_of_regparams

class RegularizedKernelRegression:
    def __init__(self, kernel):
        """
        Creates a  RegularizedKernelRegression object.
        
        The regularization parameter best suited to given data is selected
        from a user-specific list of regularization parameters via k-fold
        cross-validation.
        
        Parameters:
            kernel - function object which takes two array-like arguments X and Y
        """
        self.kernel = kernel
        self.is_fitted = False
        
    def fit(self, X, y, cv = True, k='LOOCV', edf = 3):
        """
        Fits the given N data.
        
        If cv == True, the regularization parameter is estimated using
        k-fold cross-validation. By default, LOOCV is used (k=N). If N is not a
        multiple of k, then an ValueError is raised. If cv == False,
        the user-provided regularization parameter regp is used.
        
        Parameters:
            X    - array-like with shape (N,); sample points 
            y    - array-like with shape (N,); sample values
            cv   - boolean; indicates if the regularization parameter should
                   be estimated using cross-validation.
            k    - int; the number validation sets, k=N leads to LOOCV
            regp - user-specified regularization parameter that is used if
                   cross-validation is turned off.
        
        """
        self.X_train = X
        n = np.size(y)
        
        if k == 'LOOCV':
            k = n
        
        K = self.kernel(X,X)
        eigval, Q = eigenvalue_decomposition(K)
        self.potential_edfs, self.potential_regparams = effective_degrees_of_freedom(eigval, Q)
        if cv: # use cross-validation
            if n % k != 0:
                raise ValueError('the number of training data must be a multiple of the cross-validation batch size k')
            self.cv_l2errors = test_edfs_cv(eigval, Q, y, self.potential_edfs, self.potential_regparams, k)
            best = np.argmin(self.cv_l2errors)
            self.fitted_edf = self.potential_edfs[best]
            self.fitted_regparam = self.potential_regparams[best]
        else:
            self.fitted_edf = edf
            self.fitted_regparam = self.potential_regparams[edf-3]
            
        self.fitted_coefficients = y*Q*np.diag(1/(self.fitted_regparam+eigval))*Q.transpose()
        self.is_fitted = True    
            
    def predict(self, X):
        """
        Predicts function values at the given points using the computed fit.
        
        Parameters:
            X - array-like with shape (M,)            
        Returns:
            y_pred - array-like with shape (M,); the estimated function values
        """
        y_pred = self.fitted_coefficients*self.kernel(self.X_train, X)
        return np.squeeze(np.asarray(y_pred))

    def __call__(self, X):
        """
        Simple wrapper for self.predict.

        This allows to do the following:
	pred = RegularizedKernelRegression(some_kernel)
        pred.fit(X,y)
        pred(X_) # has the same effect has invoking pred.predict(X_)
	"""
        return self.predict(X)

class GaussianProcess:
    def __init__(self, length_scale, noise_level):
        """
        Creates an instance of the GaussianProcess.
        
        Parameters:
            kernel - function object which takes two array-like arguments X and Y
            noise_level - float; the noise level
        """
        self.kernel = lambda X,Y: gauss_kernel(X,Y, 1/(2*np.power(length_scale,2)))
        self.length_scale = length_scale
        self.noise_level = noise_level
        self.is_fitted = False

    def fit(self, X,y):
        """
        Fits the Gaussian process to the given N observations
        
        Parameters:
            X - array-like with shape (N,d) or (N,) if d==1
            y - array-like with shape (N,)
        """
        N = np.size(X)
        self.X_train = X
        self.y_train = y
        self.chol = cholesky(self.kernel(X,X) + self.noise_level*np.eye(N), lower=True)
        self.y_chol = np.asmatrix(solve_triangular(self.chol, y, lower=True)).transpose()
        self.is_fitted = True
        
    def predict(self, X):
        """
        Predicts the Gaussian process at the given points X.
        
        If the Gaussian process has not been fitted to data yet,
        mean and covariance of the prior in the points X are returned
        
        Parameters:
            X - array-like with shape (M,d) or (M,) if d==1
            
        Returns:
            mu - matrix with shape (M,d); posterior mean in the points X
            sig - matrix with shape (M,M); posterior covariance in the points X
        """
        if self.is_fitted:
            k = self.kernel(self.X_train, X)
            k_chol = np.asmatrix(solve_triangular(self.chol, k, lower=True))

	    # mean and covariance of the posterior distribution
            mu  = k_chol.transpose()*self.y_chol
            sig = self.kernel(X,X) - k_chol.transpose()*k_chol
            #sig = self.kernel(X,X) - k_chol.transpose()*k_chol + self.noise_level*np.eye(np.size(X))
#	    lml = -0.5*np.asmatrix(self.y_train)*self.y_chol - np.sum(np.log(np.diag(self.chol)))-np.size(self.y_train)
        else:
            mu = np.zeros(np.size(X))
            sig = self.noise_level*np.eye(np.size(X))
        return mu, sig
    
    def generate_sample_path(self, X, n=1):
        """
        Generates sample paths of the posterior distribution
        computed in the points X.
        
        If the Gaussian process has not been fitted to any data yet,
        then sample paths of the prior are generated.
        
        To assure positive definiteness of the posterior covariance,
        0.000001 noise is added to the posterior covariance.
        
        Parameters:
            X - array-like with shape (M,)
            n - number of sample paths (default n=1)
            
        Returns:
            f - array with shape (M,n) or (M,) if n==1
        """
        M = np.size(X)
        U = np.asmatrix(np.random.standard_normal(size=(M,n)))
        if self.is_fitted:
            mu, sig = self.predict(X)
            try:
                L = cholesky(sig, lower=True)
            except np.linalg.LinAlgError:
                L = cholesky(sig+0.000001*np.eye(M), lower=True)
            return mu + L*U
        else:
            return U
    
    def plot_sample_paths(self, ax, X, n=1):
        """
        Adds sample paths of the Gaussian process to the plot.
        
        If the Gaussian process has been fitted to data, then
        realizations of the posterior distribution in the points X
        are generated, otherwise realizations of the prior distribution
        in X.
        
        Parameters:
            X - array-like with shape (M,)
            n - number of sample paths (default n=1)
        """
        Y = self.generate_sample_path(X,n)
        for i in range(n):
            ax.plot(X,Y[:,i],color='black')
            
        
    def plot_prediction(self, ax, X, legend=False):
        """
        Adds a plot to the given axis object which visualizes
        the posterior distribution computed in the points X.
        
        If the Gaussian process has not been fitted to any data yet,
        then the prior distribution is visualized.
        
        The points X are depicted as blue dots, the posterior mean
        is depicted as a red line and the 95%-confindence region is
        depicted as filled area around the posterior mean with color lightgray.
        As title the length scale and the noise level are printed.
        
        Parameters:
            X      - array-like with shape (M,)
            ax     - axis object
            legend - if True a legend is depicted in the plot
        """
        mu, sig = self.predict(X)
        mu = np.squeeze(np.asarray(mu))
        
        ax.set_title('w = '+str(self.length_scale)+'; sigma^2 = '+str(self.noise_level))
        ax.fill_between(X, mu+1.9600*np.sqrt(np.diag(sig)), mu-1.9600*np.sqrt(np.diag(sig)), color='lightgray', label='95% confidence region')
        ax.plot(X, mu, 'r', label='posterior mean') # plot posterior mean
        if self.is_fitted:
            ax.plot(self.X_train, self.y_train, 'xb', label='training data') # plot training data
        if legend:
            ax.legend(loc='upper left')

###########################################################################
# routines for anlysis
###########################################################################

def test_edfs_cv(eigval, Q, y, list_of_edfs, list_of_regparams, k):
    """
    Estimates l2 prediction errors for the given effective degrees of freedom (edf)
    and the corresponding regularization parameters using k-fold cross validation.
        
    Parameters:
        eigval            - array-like with N elements; the eigenvalues
        Q                 - matrix of shape (N,N); the normalized (unit “length”) eigenvectors,
                            such that the column Q[:,i] is the eigenvector corresponding
                            to the eigenvalue eigval[i].
        y                 - array-like with shape (N,); sample values
        list_of_edfs      - potential edfs computed by effective_degrees_of_freedom
        list_of_regparams - corresponding regularization parameters computed by effective_degrees_of_freedom
        k                 - int; the number validation sets, k=N leads to LOOCV
        
    Returns:
        list_of_l2errs    - list of l2-errors computed via cross-validation
    """
    n = np.size(y) # number of training values
    shuffled = np.random.permutation(np.arange(n)) # randomly permute training data
    
    list_of_l2errs = []
    for regp in list_of_regparams:
        D= np.diag(eigval/(regp+eigval))
        y_pred = y*Q*D*Q.transpose() # function values estimated on training data
            
        #compute estimates for validation sets
        y_cv = np.zeros(n)
        #shuffled = np.random.permutation(np.arange(n)) # randomly permute training data
        subset_selector = np.arange(n//k)
        for i in range(k):
            validation_set = shuffled[i*n//k+subset_selector]
            M = Q[validation_set,:]*D*Q[validation_set,:].transpose()
            y_cv[validation_set] = scipy.linalg.solve( np.eye(n//k) - M, np.transpose(y_pred[0,validation_set] - y[validation_set]*M))
            
        list_of_l2errs.append(np.sum(np.power(y-y_cv,2.0))/n)

    return list_of_l2errs

def test_edfs_aic(eigval, Q, y, list_of_edfs, list_of_regparams, k):
    n = np.size(y) # number of training values
    
    list_of_aics = []
    for i in range(np.size(list_of_edfs)):
        regp = list_of_regparams[i]
        edf  = list_of_edfs[i]
        D= np.diag(eigval/(regp+eigval))
        y_pred = y*Q*D*Q.transpose() # function values estimated on training data
            
        list_of_aics.append(np.sum(np.power(y-y_pred,2.0))/n + 2*edf/n*np.power(0.5,2))

    return list_of_aics


def test_error(f,g,X):
    return (1.0/np.size(X))*np.sum(np.power(f(X) - g(X),2.0))

def plot_rms_over_regparam(ax, X_test, X_sample, y, y_noisy, regparams, kernel, kernel_label):
    global toy_noisy, toy_observed
    eigval, Q = eigenvalue_decomposition(kernel(X_sample, X_sample))
    
    ax.set_title(kernel_label)
    ax.set_xlabel('Regularization parameter')
    ax.set_ylabel('Mean square error')
    
    l2_errors = []
    for regparam in regparams:
        predictor = compute_predictor( kernel, eigval, Q, X_sample, y, regparam)
        l2_errors.append(test_error(toy, predictor,X_test))
    ax.plot(regparams, l2_errors, label = 'uncorrupted')
    
    l2_errors = []
    for regparam in regparams:
        predictor = compute_predictor( kernel, eigval, Q, X_sample, y_noisy, regparam)
        l2_errors.append(test_error(toy, predictor,X_test))
    ax.plot(regparams, l2_errors, label = 'noisy')
    
    ax.legend()
